;FILE "eforth.inc" for "eforth.asm"
;-
;- ===============================================================
;- eForth by Ting 1990, Muench, Rideau, Hohensee 1999,
;- mod. for nasm + asmutils-0.14+, hp'20010323 - phpr@snafu.de
;- ===============================================================
;-

; nasm specific, no pre-requisites
%include "nasmmac.inc"
%include "reloc.inc"
list off;[nolist]
;-
;- NAMING CONVENTIONS
;-      ' is pronounced <tick> in the Forth world, alone or in words that
;-              contain it. eforth words like 'boot are execution vectors
;-              that allow redefining words "after" other words that use
;-              them.                                                     
;-      ! is pronounced "store", and means put second_on_stack into 
;-              memory at address in TOS (top of stack).           
;-      @ is pronounced "fetch", get TOS from address in TOS.        
;-      > sometimes implies "less than" and sometimes "to".        
;-	
;- ===============================================================
;-
%define __AU_DBG	; ca. 1K of additional code
;%undef  __AU_DBG

%ifndef _nolist4
%define  _nolist4 .nolist
%endif

;-	
;- ===============================================================
;-

;-
;- asmutils
list on
list on ;[list]
%include "system.inc"
%ifndef __I386_ERRNO
%include "errno.inc"
%undef __I386_ERRNO
%endif
;-
list on
list on ;[list]
;- basic defn's, conditional signal handlers setup re below
%ifndef SIGRTMIN	; not in asmutils (..0.14)
%include "linux.inc"
%elifndef SYS_rt_sigaction
%include "linux.inc"
%endif
;
; E4PN 1-3
%macro E4PN 1-3_nolist
    pushfd
    pushad
    mov eax,%{1}
%if %0>2
    mov edx,%{3}
    %ifidn %{2},s
    call e4pn.s
    %else
    call e4pn
    %endif
%else
    xor ebx,ebx
    call e4pn.n
%endif
    popad
    popfd
%endm
;- object file format dependent stuff: ELF
;- PAGE_SHIFT, basic value for PAGE_SIZE related items, dft 12, adjust tu suit.
;OS dependent constant, cannot be determined from running host system
%assign PAGE_SHIFT 12
%assign PAGE_SIZE  (1<<PAGE_SHIFT)
%assign PAGE_MASK  (-PAGE_SIZE)
%assign K1	   1024
;- pg(x)	for page size adjustment, takes (x)K bytes
%define pg(x) (((x)*1024+PAGE_SIZE-1)&PAGE_MASK)
;-	
;- ===============================================================
;-	

;-
;- Version control

%assign REL 1
%assign VER 1
%assign EXT 00
%include "version.inc"	;[hp]

;- Constants

;- cpu (ansi pedantic, not much of use in the real world...)
%assign AUB	8	; bits per address unit
%assign ACALL   0xE8909090; aligned call instr
%assign SCALL	0xE82e2e2e; seg-prefixed call
%assign ACALL   SCALL	; aligned call instr

%assign CELLX	2	; 2^CELLX = cell size
%assign CELSZ	1<<CELLX
%assign NBPW	AUB*CELSZ; #bits/word

;- word-header flags - [hp] bits, name-length set by macro
%assign widtb	5
%assign smudb	5
%assign cplob	6
%assign imedb	7
%assign NWIDTH	(1<<widtb)-1; name length maximum
%assign SMUDG	1<<smudb
%assign COMPO	1<<cplob
%assign IMEDD	1<<imedb
;%assign SMUDG  0x20
;%assign COMPO  0x40
;%assign IMEDD  0x80
;%assign MNAME  0x7F7F7F1F	US-ASCII, dreck...
%assign MNAME  0xfFfFfF1F

%assign UD	80	; disp of up-reference to u0, bot of user vari

%assign DBASE	10	; dft number conversion radix
%assign NVOCS   8	; cells, size of voc-stack
%assign NPAD	80	; bytes, pad disp from "here"
%assign NTIB	PAGE_SIZE;bytes, effective minimal size of tib

;- iso char cons
%define pchmask 0xffa07e20; hi|lo ranges of printing chars
%assign BSP	8	; backspace, i.e. printing posn 1 col. left, nothing else!
%assign LF	10	; line feed, Linux implicitely adds <CR>, thus <lf> is <nl>
%assign CR	13	; carriage return
%assign SI	15	; charset switch
%assign ESC	27
%assign BL	32	; overwriting, blank space
%assign DEL	127	; delete char at cursor posn, new printing posn 1 col. left
%assign CSI	0x9b	; <csi>

;-
;- uBITS bits
;
;- bcaps,mcaps	case-dependent if bit set
%assign bcaps 0
%assign mcaps 1<<bcaps

;-	

;-
;- Forth regs
;-	all 'all purpose'-registers used in forth code,
;-	neither sements nor special regs used or accessible,
;-	fp/mmx-regs unused, may communicate between words.
;most silly, esp for data-stack! effectively (and very efficiently) 
;inhibits kernel subroutines! -> change to: PS=edi, RS=esp
;push/pop is short but not fast, stosd/mov+lea not slower (re "lib4th")!
;-	 REGISTER	USEAGE
;-	eax	T	auxilary
%define T	eax
;-	ebx	@TOS	top item of data-stack
%define @TOS	ebx
;-	ecx	W	word address, aux.
%define W	ecx
;-	edx	T2	auxilary
%define T2	edx
;-	esi	IP	instruction ptr
%define IP	esi
;-	edi	UR	ref.reg to user-vari (cmove,fill)
%define UR	edi
%define dref	"set in e4th.asm"
%define iref	"set in e4th.asm"
%define uref byte edi-80+
;-	ebp	RS	return-stack
%define RS	ebp
%define s_RS    s_ebp
;-	%define @TOR	[RS]
%define @TOR	[RS]
%define TOR(x)	[RS+((x)*CELSZ)]
;-	%macro rpop 0-*,@TOS
;-		rpop		pop @RS and copy to @TOS
;-		rpop reg[,reg,...]
;-		rpop regl	pop from RS to 'regl' - re <rpush>
;-		rpop {num}	pop num times @RS to @TOS
;-		rpop <ea>	pop @RS to non-reg <ea>
%macro rpop 0-*.nolist
%xdefine %%l _listing_
%if %{0}=0
list %%l
    mov @TOS,dword[RS]
    lea RS,[RS+CELSZ]
list off;[nolist]
%elif %{0}=1
    %ifnum %{1}
list %%l
    mov @TOS,[RS+(%{1}*CELSZ)-CELSZ]
    lea RS,[RS+(%{1}*CELSZ)]
list off;[nolist]
    %else
	%ifid %{1}
list %%l
    mov %{1},dword[RS]
list off;[nolist]
	%else
list %%l
    push eax
    mov eax,[RS]
    mov %{1},eax
    pop eax
list off;[nolist]
	%endif
list %%l
    lea RS,[RS+CELSZ]
list off;[nolist]
    %endif
%else;%0>1
    %assign %%n %{0}
    %assign %%i %{0}+1
    %rep %{0}
    %rotate -1
    %assign %%i %%i-1
list %%l
    mov %{1},dword[RS+((%%n-%%i)*CELSZ)]
list off;[nolist]
    %endrep
list %%l
    lea RS,[RS+(%{0}*CELSZ)]
list off;[nolist]
%endif
list %%l
%xdefine rregl empty
%endm
;
;-	%macro rpush 0-*,@TOS
;-	pre-decrement & store to RS, %define <regl> to list of args
;-		rpush		copy @TOS to RS
;-		rpush reg[,reg,...]	copy reglist to RS, last arg to TOR(0)
;-		rpush {num}	copy num times @TOS to RS
;-		rpush <ea>	copy non-reg ea
%macro rpush 0-*.nolist
%xdefine %%l _listing_
list off;[nolist]
    %undef rregl
%if %{0}=0
list %%l
    lea RS,[RS-CELSZ]
    mov dword[RS],@TOS
list off;[nolist]
%xdefine rregl @TOS
%elif %{0}=1
    %ifnum %{1}
    %assign %%n %{1}
list %%l
    lea RS,[RS-(%{1}*CELSZ)]
list off;[nolist]
    %rep %%n
    %rotate -1
    %assign %%n %%n-1
list %%l
    mov dword[RS+(%%n*CELSZ)-CELSZ],@TOS
list off;[nolist]
	%ifdef rregl
    %xdefine rregl rregl,@TOS
	%else
    %xdefine rregl %{1}
	%endif
    %endrep
    %else
list %%l
    lea RS,[RS-CELSZ]
list off;[nolist]
	%ifid %{1}
list %%l
    mov dword[RS],%{1}
list off;[nolist]
	%else
list %%l
    push eax
    mov eax,%{1}
    mov [RS],eax
    pop eax
list off;[nolist]
	%endif
    %xdefine rregl %{1}
    %endif
%else
    %assign %%n %{0}
    %assign %%i %{0}+1
list %%l
    lea RS,[RS-(%%n*CELSZ)]
list off;[nolist]
    %rep %%n
    %rotate -1
    %assign %%i %%i-1
list %%l
    mov dword[RS+((%%n-%%i)*CELSZ)],%{1}
list off;[nolist]
    %ifdef rregl
    %xdefine rregl %{1},rregl
    %else
    %xdefine rregl %{1}
    %endif
    %endrep
%endif
list %%l
%endm
;-	esp	PS, beginning w/ 2nd item, @tos = ebx
%define PS	esp
%define s_PS    s_esp
;-	%define TOS(x) [esp+(((x)-1)*CELSZ)]
%define TOS(x)	[esp+(((x)-1)*CELSZ)]
;-	%macro spop 0-1_nolist4
%macro spop 0-1_nolist4
%if %0>0
    %ifnum %{1}
	%if %{1}>1
    lea esp,[esp+4*%{1}-4]
	%endif
    pop ebx
    %endif
    %ifid %{1}
    pop %{1}
    %else
	%ifnnum %{1}
%xdefine x %{1}
%error "spop x"
	%endif
    %endif
%else
    pop ebx
%endif
%endm
;-	%macro spush 0-1_nolist4
;-	pre-decrement & store to PS
%macro spush 0-1_nolist4
%if %0>0
    %ifnum %{1}
	%if %{1}>1
    lea esp,[esp-4*%{1}]
    mov [esp],ebx
    mov [esp+4*%{1}-4],ebx
	%else
    push ebx
	%endif
    %elifid %{1}
    _push %{1}
    %else
%xdefine x %{1}
%error "spop x"
    %endif
%else
    push ebx
%endif
%endm
;-	mmX	unused
;-	stX	unused
;-

;-
;- -- memory set-up --
;-
;- DRTOP  -> top=namespace to R0
%assign NM_SZ	pg(64)
;- TIBB	-> R0 to TIB
%assign RS_SZ   pg(64)
;- UPP	-> TIB to UP
%assign TIB_SZ  pg(NTIB/K1)
;- DSTOP	-> UP to 
%assign US_SZ   pg(64)
;-	-> DS to data-/code-space
%assign SS_SZ   pg(64)
;-	-> cpld'd code to cpld data
%assign DS_SZ	pg(64)
;-	-> cpl'd data to "here"
%assign CS_SZ	pg(64)
;-
;- compilespace & data in .bss,
;-  [ r/x kernel | r/w kernel-data | r/w/x dictionary ) ->	.bss	]
;- system in .ss sections, for expandability w. sys_brk;
;-  [	.ss 	<- ( wordheaders | ps | tib | uvari | rs | .ss ]
%assign SM SS_SZ+US_SZ+TIB_SZ+NM_SZ	; system-mem	towards lower addr ( headers | ps | |
%assign CM DS_SZ+CS_SZ			; compile-mem	[ data & code ) extends into .bss
;-
%define calign align 4
;- __DATASEG__ 
;kernel words' names cannot be stored to data-seg - else segv!
%macro __DATASEG__ 0_nolist4
%ifnidn __QSEG__,_DATASEG_
%ifnidn __SEG4__,.text
	section .text
%xdefine __SEG4__ .text
%endif
%define __align align CELSZ,db 0
%xdefine __QSEG__ _DATASEG_
%endif
%endm
;- __CODESEG_
%macro __CODESEG__ 0_nolist4
%ifnidn __QSEG__,_CODESEG_
%ifnidn __SEG4__,.text
	section .text
%xdefine __SEG4__ .text
%endif
%define __align align CELSZ,db 0x90
%xdefine __QSEG__ _CODESEG_
%endif
%endm
;-

;-  Define assembly macros

;[hp'20010328]
;- stng [len],string,alignval,aligncode
;-	if non-empty arg <len> passed cpl byte <len> and text, wo. alignment, else
;-	cpl byte-counted txt and <aligncode>-padded  align to <alignval> bounds
;-	string may be enclosed in '"' or an idientifier ("nasm" terminology)
%macro stng 2-4_nolist 4,0
%ifid %{00}
%{00}:
%endif
%ifid %{2}
%xdefine %%s "%{2}"
%else
%xdefine %%s %{2}
%endif
%ifnum %{1}
%xdefine %%l
	db %{1}
%else
%xdefine %%l align %{3},db %{4}
	db %%e-%%a
%endif
%%a:
	db %%s
%%e:
	%%l
%endm
;- %macro __CODE 3-4_nolist4	;lex, name, label [,wordlist]
;-     Compile a code definition header.
;-[hp]	name length not required, "lex" may be ((compile-bits)|(non-zero))
;-	zero name-length is flag to compiling a headerless word
;-	%4: this word's wordlist
;-	word-header:
;-	 code:	N[ link | L{name}.b | "name" 4-aligned,0
;-		C[ 4-aligned,nop {name}:
;-	 colon:	N[ link | Lc{name}.b | "name" 4-aligned,0
;-		C[ 4-aligned,nop c{name}: | 2e,2e,2e,call dolist, {name}:
;-	 vocab:	N[ link | Lc{name}.b | "name" 4-aligned,0
;-		C[ 4-aligned,nop c{name}: | 2e,2e,2e, {name}: call dolist, dovoc, vhead, vlink
;		%xdefine _nolist4
%macro __CODE 3-4_nolist4 FTH
        __CODESEG__
        calign
%if %{1} = 0
%{3}:
%else
	__DATASEG__
	__align
	dd %{3}		; cfa
	dd %{4}NAME	; lfa
Y.%{3}:			; nfa
	db (%%e-%%a)|((%{1})&224)
%%a:
	db %{2}
%%e:
	__align
	__CODESEG__
    %if %{1}>=0
%{3}:
    %endif
%xdefine %{4}CODE %{3}	; xt toplink -> CODE%{4}
%xdefine %{4}NAME Y.%{3}; na linkage -> NAME%{4}
%xdefine _NAME Y.%{3}
%endif
%xdefine _%{3}_ %{3}
%endm

%define __NCODE(x,y) __CODE 0,x,y

;- %macro __NCOLON 3_nolist4
;-	[hp] headerless colon defn
;-	local label <.n> at hi-level code entry
%xdefine _nolist4 .nolist
%macro __NCOLON 3_nolist4
	__CODESEG__
	calign;;__align
	db 0x2E,0x2E,0x2E
%{3}:
	call DOLST
.n:
%endm

;- %macro __COLON 3_nolist4	;(lex, name, label)
;-	Compile a colon definition header.
;-	Note that the call to DOLST must be padded out to a CELSZ boundary;
;-	a call takes 5 bytes so we add in segment override prefixes (useless 
;-	ones, at that - [hp: now un-used...]) to pad out the size.
; delayed entry label generation, to saveing cpu from executing filler bytes
;		%xdefine _nolist4
%macro __COLON 3-4_nolist4 FTH
	__CODE -65536|%{1},%{2},%{3},%{4}
	db 0x2E,0x2E,0x2E
%{3}:
	call DOLST
.nt:
%endm

;- label: %macro __AWORD 3_nolist4
;-	alias code or colon defn, <label> required!
%xdefine _nolist4 .nolist
%macro __AWORD 3-4_nolist4 FTH
	__CODE %{1},%{2},%{00},%{4}
%ifidn _%{3}_,%{3}
	jmp disp %{3}
%else
	jr %{3}
%endif
%endm

;- %macro __USER 3_nolist4		;(lex, name, label)
;-	Compile a user variable header.
;-	[hp] generates label <u%{3} equ _USER> for [uref...] addressing
;-	     <%define p_user> for fast code generation of kernel uvari
%macro __USER 3-4_nolist4 FTH
    __CODE %{1},%{2},%{3},%{4}
u%{3} equ _USER
%ifdef p_user
	push ebx
	__ENTER
	lea ebx,[uref u%{3}]
	__NEXT
%else
	calign
	db 0x2E,0x2E,0x2E
	call DOLST
	dd DOUSE,_USER
%endif
%assign _USER _USER+CELSZ
%endm

;- %macro __CONS 4_nolist4
;-	define a literal value
;-	+20 bytes ea., saves kernel space after 5 coccurrences
%macro __CONS 4-5_nolist4 FTH
    __CODE %{1},%{2},%{3},%{5}
    __ENTER
    push ebx
    _mov ebx,%{4}
    __NEXT
%endm

;-
 %define __ENTER lea IP,[IP+CELSZ]
;%macro __ENTER 0-1 0;_nolist4 0
;    lea IP,[IP+((%{1})+1)*CELSZ]
;%endm
;- %define __NEXT jmp [esi]
%define __NEXT jmp [IP]
;- %define __NEXT1 push to esp (=PS!) ret-addr to <next1>, at __ENTER w/ __NEXT
%define __NEXT1 push dword next1
;- %define __NEXT2 push to esp (=PS!) ret-addr to <next2>, at __NEXT
%define __NEXT2 push dword next2
;
%ifid __NASM,9821
%imacro __EXIT 0-*.nolist
%die '__EXIT -> __NEXT -?-'
%endm
%else
%imacro __EXIT 0-*.nolist
%error '__EXIT -> __NEXT -?-'
%endm
%endif
;-
;- Initialize assembly variable
;
;-
;- %define lit(x) DOLIT,x	hi-level literal
%define lit(x) DOLIT,x
;-
;%define def_vocs
; --> :segv: all ptr combinations tried, no chance(?), requires cpl. re-design -?-
%undef  def_vocs
;
%assign  FTHCODE 0	; names linkage, "forth"
%xdefine FTHNAME vocbase.y; names linkage, "forth"
%ifdef def_vocs
%assign  ROOCODE 0	; "root" vocabulary
%assign  SYSCODE 0	; "system" vocabulary
;
%xdefine ROONAME vocbase.y; "root" vocabulary
%xdefine SYSNAME vocbase.y; "system" vocabulary
%endif
;-	

;-
;- conditional system setup
;
%define vt_ini		; 100; single-key mode
%define sig_hnd		; 220; rst console mode on exit, etc.
;%define sys_calls	;5260; include 190+ named syscalls...
			;    ; ...not really required - re generic 'syscall'
; extra sighnd', '%undef'-ine in asm-file, if required:
%define hup_jobs	;  32; terminate child jobs when self being terminated
%define fpe_sig		;  60; activate div0 and ovf handler
%define segv_sig	;    ; recovering from a single fault
%define ileg_sig	;    ; recovering from a single fault
%define winch_sig	;    ; - n.i. -

; sighnd install masks
%assign fpesig   00000000000000000000000100000000b; use special or, parent's hnd
%assign winchsig 00010000000000000000000000000000b
%assign segvsig  00000000000000000000100010000000b; segv & buserror
%assign ilegsig  00000000000000000000000000010000b
%assign qitsig   00000000000000000000000000000100b
%ifndef fpe_sig
%assign fpesig 0
%endif
%ifndef winch_sig
%assign winchsig 0
%endif
%ifndef segv_sig
%assign segvsig 0
%endif
%ifndef ileg_sig
%assign segvsig 0
%endif
%ifndef quit_sig
%assign quitsig 0
%endif

%undef nonsense	; de-activate silly or just useless words...

;- ===============================================================
; eforth.inc <eof>
